#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include "array.h"

#include "dijkstra.h"

dijkstra_map *
dijkstra_map_new(int size_x, int size_y)
{
  dijkstra_map *p = NULL;

  if (size_x <= 0)
  {
    fprintf(stderr, "dijkstra_map_new: size_x is non-positive (%d)\n",
            size_x);
    return NULL;
  }
  if (size_y <= 0)
  {
    fprintf(stderr, "dijkstra_map_new: size_y is non-positive (%d)\n",
            size_y);
    return NULL;
  }
  if (size_x * size_y <= 0)
  {
    fprintf(stderr, "dijkstra_map_new: size %d x %d is too big\n",
            size_x, size_y);
    return NULL;
  }

  p = (dijkstra_map *) malloc(sizeof(dijkstra_map));
  if (p == NULL)
  {
    fprintf(stderr, "dijkstra_map_new: malloc failed\n");
    return NULL;
  }

  p->size_x = 0;
  p->size_y = 0;
  p->result = NULL;
  p->length = 0;
  p->num_queue = 0;
  p->queue = NULL;
  p->num_queue_next = 0;
  p->queue_next = NULL;

  p->size_x = size_x;
  p->size_y = size_y;

  p->result = array2_new(p->size_x, p->size_y);
  if (p->result == NULL)
  {
    fprintf(stderr, "dijkstra_map_new: array2_new(result) failed\n");
    dijkstra_map_delete(p);
    p = NULL;
    return NULL;
  }

  p->queue = array2_new(p->size_x * p->size_y, 2);
  if (p->queue == NULL)
  {
    fprintf(stderr, "dijkstra_map_new: array2_new(queue) failed\n");
    dijkstra_map_delete(p);
    p = NULL;
    return NULL;
  }

  p->queue_next = array2_new(p->size_x * p->size_y, 2);
  if (p->queue_next == NULL)
  {
    fprintf(stderr, "dijkstra_map_new: array2_new(queue_next) failed\n");
    dijkstra_map_delete(p);
    p = NULL;
    return NULL;
  }

  return p;
}

void
dijkstra_map_delete(dijkstra_map *p)
{
  if (p == NULL)
    return;

  if (p->result != NULL)
  {
    array2_delete(p->result, p->size_x, p->size_y);
    p->result = NULL;
  }

  if (p->queue != NULL)
  {
    array2_delete(p->queue, p->size_x * p->size_y, 2);
    p->queue = NULL;
  }
  p->num_queue = 0;

  if (p->queue_next != NULL)
  {
    array2_delete(p->queue_next, p->size_x * p->size_y, 2);
    p->queue_next = NULL;
  }
  p->num_queue_next = 0;

  p->size_x = 0;
  p->size_y = 0;

  free(p);
  p = NULL;
}

void
dijkstra_map_reset(dijkstra_map *p)
{
  int x;
  int y;

  if (p == NULL)
  {
    fprintf(stderr, "dijkstra_map_reset: p is NULL\n");
    return;
  }

  for (x = 0; x < p->size_x; x++)
  {
    for (y = 0; y < p->size_y; y++)
    {
      p->result[x][y] = -1;
    }
  }

  p->length = 0;
  p->num_queue = 0;
  p->num_queue_next = 0;
}

/* this function does _not_ reset p (arg 1) */
int
dijkstra_map_init(dijkstra_map *p, int x, int y)
{
  if (p == NULL)
  {
    fprintf(stderr, "dijkstra_map_init: p is NULL\n");
    return 1;
  }

  p->result[x][y] = p->length;
  p->queue[p->num_queue][0] = x;
  p->queue[p->num_queue][1] = y;
  (p->num_queue)++;

  return 0;
}

/* map_blocked (arg 2) must be an array of size (p->size_x, p->size_y)
 * map_blocked[x][y] must be non-zero if and only if
 * the grid (x, y) is blocked
 */
int
dijkstra_map_refresh(dijkstra_map *p, int **map_blocked, int max_length)
{
  int i;
  int j;
  int x;
  int y;
  int dx;
  int dy;
  int temp;
  int **queue_temp = NULL;

  if (p == NULL)
  {
    fprintf(stderr, "dijkstra_map_refresh: p is NULL\n");
    return 1;
  }
  if (map_blocked == NULL)
  {
    fprintf(stderr, "dijkstra_map_refresh: map_blocked is NULL\n");
    return 1;
  }

  while (p->num_queue > 0)
  {
    if ((max_length >= 0) && (p->length > max_length))
      break;

    p->num_queue_next = 0;

    for (i = 0; i < p->num_queue; i++)
    {
      for (j = 0; j < 9; j++)
      {
        if (j == 4)
          continue;

        dx = -1 + (j % 3);
        dy = 1 - (j / 3);

        x = p->queue[i][0] + dx;
        y = p->queue[i][1] + dy;

        if ((x < 0) || (x >= p->size_x))
          continue;
        if ((y < 0) || (y >= p->size_y))
          continue;
        if (map_blocked[x][y])
          continue;
        if (p->result[x][y] >= 0)
          continue;

        p->result[x][y] = p->length + 1;
        p->queue_next[p->num_queue_next][0] = x;
        p->queue_next[p->num_queue_next][1] = y;
        (p->num_queue_next)++;
      }
      ;
    }

    temp = p->num_queue;
    p->num_queue = p->num_queue_next;
    p->num_queue_next = temp;

    queue_temp = p->queue;
    p->queue = p->queue_next;
    p->queue_next = queue_temp;

    (p->length)++;
  }

  return 0;
}
